/*
  Tutanchamun, an arcade game
  Copyright (C) 1998  David Kastrup
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
/* This is the graphics interface of Tutanchamun.  The tasks it mainly
 * has to do is initialization of graphics, putting things to screen,
 * scrolling, etc.  Note that getting a bitmap back from the screen is
 * typically a costly operation, so it is a good idea to work on some
 * kind of client memory pixmaps and batch-copy the data to screen.
 * For X this is probably the best method of operation, anyway, as
 * reading back things from the screen forces a synchronization with
 * the X server, and other alternatives of plotting masked shapes need
 * to employ two plotting operations with logical operations and
 * possibly causing flicker.  The interface routines here also convert
 * character maps to pictures, whatever those pictures happen to be.
 */

#define PLAYGROUNDX 2
#define PLAYGROUNDY 8

#include "tutdisp.h"
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <vga.h>
#include <vgakeyboard.h>
#include <vgagl.h>
#include <assert.h>

unsigned char playground[12*16][14*16];

void *d_makefgshape(Tile tile, Mask mask, int reflected)
{
  int di, lastx, x, y, size;
  unsigned char *dst;
  int warp = reflected ? 0 : 15;
  size = 0;
  for (y=0; y<16; y++) {
    x=0;
    for (;;) {
      lastx = x;
      for (;x<16 && !((mask[y] >> (x^warp)) & 1); x++)
	;
      size++;
      if (x==16)
	break;
      lastx = x;
      for (;x<16 && ((mask[y] >> (x^warp)) & 1); x++)
	;
      size += x-lastx + 1;
    }
  }
  
  dst = malloc(size);
  assert(dst != NULL);

  di = 0;
  for (y=0; y<16; y++) {
    x=0;
    for (;;) {
      lastx = x;
      for (;x<16 && !((mask[y] >> (x^warp)) & 1); x++)
	;
      dst[di++] = x - lastx;
      if (x==16)
	break;
      lastx = x;
      for (;x<16 && ((mask[y] >> (x^warp)) & 1); x++) {
	dst[di+x+1-lastx] =
	  4*(tile[0][y] >> (x^warp) & 1)
	  + 2* (tile[1][y] >> (x^warp) & 1)
	  + (tile[2][y] >> (x^warp) & 1);
      }
      dst[di] = x-lastx;
      di += x-lastx + 1;
    }
  }
  assert(di == size);
  return dst;
} 

void d_redrawtile(int x0, int y0)
{
  int xs;

  if (x0 < 0) {
    assert(x0 > -4);
    xs = x0+4;
    x0 = 0;
  } else if (x0 > 4*12-4) {
    assert(x0 < 4*12);
    xs = 4*12 - x0;
  } else
    xs = 4;
  gl_putboxpart(PLAYGROUNDX+4*x0, PLAYGROUNDY+4*y0, xs*4, 16,
		16*14, 16*12, playground, 4*x0+16, 4*y0);
}



void d_fgshape(void *p, int x0, int y0)
{
  int x, y, si;
  unsigned char *src = p;

  si = 0;
  for (y=0; y<16; y++) {
    for (x=0; (x += src[si])<16;) {
      memcpy(&playground[y0*4+y][x0*4+x+16], &src[si+2], src[si+1]);
      x += src[si+1];
      si += src[si+1]+2;
    }
    ++si;
  }
  d_redrawtile(x0,y0);
}

void d_remshape(void *p, int x0, int y0, int colour)
{
  int x, y, si;
  unsigned char *src = p;

  si = 0;

  for (y=0; y<16; y++) {
    for (x=0; (x += src[si])<16;) {
      memset(&playground[y0*4+y][x0*4+x+16], colour, src[si+1]);
      x += src[si+1];
      si += src[si+1]+2;
    }
    ++si;
  }
  d_redrawtile(x0,y0);
}
      
void *d_makebgshape(Tile p)
{
  char *res;
  int x,y;

  res = malloc(256);
  assert(res != NULL);
  for (y=0; y<16; ++y) {
    for (x=0; x<16; ++x) {
      res[(x^15)+y*16] = (4*(p[0][y]>>x&1) | 2*(p[1][y]>>x&1) |
	  (p[2][y]>>x&1));
    }
  }
  return res;
}

void d_bgshape(void *p, int x, int y)
{
  unsigned char (*pb)[16] = p;
  int i;

  gl_putbox(PLAYGROUNDX+4*x, PLAYGROUNDY+4*y, 16, 16, p);
  for (i=0; i<16; ++i) {
    memcpy(&playground[y*4+i][x*4+16], pb[i], 16);
  }
}
  

void *d_makepicshape(unsigned short p[3][48][3])
{
  char *res;
  unsigned x,y;

  res = malloc(48*48);
  assert(res != NULL);
  for (y=0; y<48; ++y) {
    for (x=0; x<48; ++x) {
      res[(x^15)+y*48] = ( 4*(p[0][y][x/16]>>(x%16)&1)
	                 | 2*(p[1][y][x/16]>>(x%16)&1)
	                 |   (p[2][y][x/16]>>(x%16)&1));
    }
  }
  return res;
}

void *d_makemaskshape(int color, Mask *mask,
			    void *source, int orientation)
{
  int x, y;
  char (*src)[16] = malloc(256);
  assert(src != 0);
  memcpy(src, source, 16*16);
  for (y=0; y<16; y++)
    for (x=0; x<16; x++)
      if (((*mask)[y]>>(15-x))&1)
	src[y][orientation ? 15-x : x] = color;
  return src;
}

void d_destroyfgshape(void *p)
{
  free(p);
}

void d_destroybgshape(void *p)
{
  free(p);
}

void d_destroymaskshape(void *p)
{
  free(p);
}

void d_destroypicshape(void *p)
{
  free(p);
}

void d_scrlft(void *buf[12])
{
  int row, i;

  memmove(&playground[0][16], &playground[0][32], 16*16*14*12 - 48);
  for (row=0; row<12; ++row) {
    for (i=0; i<16; ++i) {
      memcpy(&playground[row*16+i][12*16], ((unsigned char (*)[16])buf[row])[i], 16);
    }
  }
  gl_putboxpart(PLAYGROUNDX, PLAYGROUNDY, 12*16, 12*16, 14*16, 12*16,
		playground, 16, 0);
}



void d_scrrig(void *buf[12])
{
  int row, i;

  memmove(&playground[0][32], &playground[0][16], 16*16*14*12 - 48);
  for (row=0; row<12; ++row) {
    for (i=0; i<16; ++i) {
      memcpy(&playground[row*16+i][16], ((unsigned char (*)[16])buf[row])[i], 16);
    }
  }
  gl_putboxpart(PLAYGROUNDX, PLAYGROUNDY, 12*16, 12*16, 14*16, 12*16,
		playground, 16, 0);
}

void d_pict(void *p, int x, int y)
{
  gl_putbox(PLAYGROUNDX+4*x, PLAYGROUNDY+4*y, 48, 48, p);
}

void d_rect(int colour, int x, int y, int xs, int ys)
{
  int i;

  gl_fillbox(PLAYGROUNDX+4*x, PLAYGROUNDY+4*y, xs*4, ys*4, colour);
  for (i=0; i<4*ys; ++i) {
    memset(&playground[y*4+i][x*4+16], colour, xs*4);
  }
}

void d_liftgates(int x, int y, int colour)
{
  int i;

  for (i=0; i<12; i++) {
    memcpy(&playground[y*4+i][x*4+16],
	   &playground[y*4+i+4][x*4+16],
	   16);
  }
  for (i=12; i<16; i++) {
    memset(&playground[y*4+i][x*4+16], colour, 16);
  }
  d_redrawtile(x,y);
}

void d_opendoors(void *pic, int x0, int y0, int phase)
{
  unsigned char (*src)[48] = pic;
  int y;

  for (y=0; y<48; y++) {
    playground[y0*4+y][x0*4+16+23-phase] = src[y][23-phase];
    playground[y0*4+y][x0*4+16+24+phase] = src[y][24+phase];
  }
  gl_putboxpart(PLAYGROUNDX+x0*4+23-phase, PLAYGROUNDY+y0*4,
		1, 48, 14*16, 12*16,
		playground, x0*4+16+23-phase, y0*4);
  gl_putboxpart(PLAYGROUNDX+x0*4+24+phase, PLAYGROUNDY+y0*4,
		1, 48, 14*16, 12*16,
		playground, x0*4+16+24+phase, y0*4);
} 

void d_drawshot(unsigned x, unsigned y, int w, int colour)
{
  assert(w > 0);
  memset(&playground[y*4][x*4+16], colour, w*4);
  gl_hline(PLAYGROUNDX+4*x, PLAYGROUNDY+4*y, PLAYGROUNDX+4*(x+w)-1, colour);
}

void d_clearmansma()
{
  gl_fillbox(PLAYGROUNDX+192, PLAYGROUNDY-8, 96, 32, 0);
}

void d_drasma(int n, void *p)
{
  if (n<5)
    gl_putbox(PLAYGROUNDX+200+16*n, PLAYGROUNDY-8, 16, 16, p);
}

void d_clearsma(int n)
{
  if (n<5)
    gl_fillbox(PLAYGROUNDX+200+16*n, PLAYGROUNDY-8, 96, 32, 0);
}

void d_draman(int n, void *p)
{
  gl_putbox(PLAYGROUNDX+200+16*n, PLAYGROUNDY+8, 16, 16, p);
}

void d_exit(void)
{
  vga_setmode(TEXT);
}

unsigned char *d_makebigfont(unsigned char *p, int size)
{
  static unsigned char conv[256][2] = {
    {0x00, 0x00},
    {0x00, 0x03},
    {0x00, 0x0c},
    {0x00, 0x0f},
    {0x00, 0x30},
    {0x00, 0x33},
    {0x00, 0x3c},
    {0x00, 0x3f},
    {0x00, 0xc0},
    {0x00, 0xc3},
    {0x00, 0xcc},
    {0x00, 0xcf},
    {0x00, 0xf0},
    {0x00, 0xf3},
    {0x00, 0xfc},
    {0x00, 0xff},
    {0x03, 0x00},
    {0x03, 0x03},
    {0x03, 0x0c},
    {0x03, 0x0f},
    {0x03, 0x30},
    {0x03, 0x33},
    {0x03, 0x3c},
    {0x03, 0x3f},
    {0x03, 0xc0},
    {0x03, 0xc3},
    {0x03, 0xcc},
    {0x03, 0xcf},
    {0x03, 0xf0},
    {0x03, 0xf3},
    {0x03, 0xfc},
    {0x03, 0xff},
    {0x0c, 0x00},
    {0x0c, 0x03},
    {0x0c, 0x0c},
    {0x0c, 0x0f},
    {0x0c, 0x30},
    {0x0c, 0x33},
    {0x0c, 0x3c},
    {0x0c, 0x3f},
    {0x0c, 0xc0},
    {0x0c, 0xc3},
    {0x0c, 0xcc},
    {0x0c, 0xcf},
    {0x0c, 0xf0},
    {0x0c, 0xf3},
    {0x0c, 0xfc},
    {0x0c, 0xff},
    {0x0f, 0x00},
    {0x0f, 0x03},
    {0x0f, 0x0c},
    {0x0f, 0x0f},
    {0x0f, 0x30},
    {0x0f, 0x33},
    {0x0f, 0x3c},
    {0x0f, 0x3f},
    {0x0f, 0xc0},
    {0x0f, 0xc3},
    {0x0f, 0xcc},
    {0x0f, 0xcf},
    {0x0f, 0xf0},
    {0x0f, 0xf3},
    {0x0f, 0xfc},
    {0x0f, 0xff},
    {0x30, 0x00},
    {0x30, 0x03},
    {0x30, 0x0c},
    {0x30, 0x0f},
    {0x30, 0x30},
    {0x30, 0x33},
    {0x30, 0x3c},
    {0x30, 0x3f},
    {0x30, 0xc0},
    {0x30, 0xc3},
    {0x30, 0xcc},
    {0x30, 0xcf},
    {0x30, 0xf0},
    {0x30, 0xf3},
    {0x30, 0xfc},
    {0x30, 0xff},
    {0x33, 0x00},
    {0x33, 0x03},
    {0x33, 0x0c},
    {0x33, 0x0f},
    {0x33, 0x30},
    {0x33, 0x33},
    {0x33, 0x3c},
    {0x33, 0x3f},
    {0x33, 0xc0},
    {0x33, 0xc3},
    {0x33, 0xcc},
    {0x33, 0xcf},
    {0x33, 0xf0},
    {0x33, 0xf3},
    {0x33, 0xfc},
    {0x33, 0xff},
    {0x3c, 0x00},
    {0x3c, 0x03},
    {0x3c, 0x0c},
    {0x3c, 0x0f},
    {0x3c, 0x30},
    {0x3c, 0x33},
    {0x3c, 0x3c},
    {0x3c, 0x3f},
    {0x3c, 0xc0},
    {0x3c, 0xc3},
    {0x3c, 0xcc},
    {0x3c, 0xcf},
    {0x3c, 0xf0},
    {0x3c, 0xf3},
    {0x3c, 0xfc},
    {0x3c, 0xff},
    {0x3f, 0x00},
    {0x3f, 0x03},
    {0x3f, 0x0c},
    {0x3f, 0x0f},
    {0x3f, 0x30},
    {0x3f, 0x33},
    {0x3f, 0x3c},
    {0x3f, 0x3f},
    {0x3f, 0xc0},
    {0x3f, 0xc3},
    {0x3f, 0xcc},
    {0x3f, 0xcf},
    {0x3f, 0xf0},
    {0x3f, 0xf3},
    {0x3f, 0xfc},
    {0x3f, 0xff},
    {0xc0, 0x00},
    {0xc0, 0x03},
    {0xc0, 0x0c},
    {0xc0, 0x0f},
    {0xc0, 0x30},
    {0xc0, 0x33},
    {0xc0, 0x3c},
    {0xc0, 0x3f},
    {0xc0, 0xc0},
    {0xc0, 0xc3},
    {0xc0, 0xcc},
    {0xc0, 0xcf},
    {0xc0, 0xf0},
    {0xc0, 0xf3},
    {0xc0, 0xfc},
    {0xc0, 0xff},
    {0xc3, 0x00},
    {0xc3, 0x03},
    {0xc3, 0x0c},
    {0xc3, 0x0f},
    {0xc3, 0x30},
    {0xc3, 0x33},
    {0xc3, 0x3c},
    {0xc3, 0x3f},
    {0xc3, 0xc0},
    {0xc3, 0xc3},
    {0xc3, 0xcc},
    {0xc3, 0xcf},
    {0xc3, 0xf0},
    {0xc3, 0xf3},
    {0xc3, 0xfc},
    {0xc3, 0xff},
    {0xcc, 0x00},
    {0xcc, 0x03},
    {0xcc, 0x0c},
    {0xcc, 0x0f},
    {0xcc, 0x30},
    {0xcc, 0x33},
    {0xcc, 0x3c},
    {0xcc, 0x3f},
    {0xcc, 0xc0},
    {0xcc, 0xc3},
    {0xcc, 0xcc},
    {0xcc, 0xcf},
    {0xcc, 0xf0},
    {0xcc, 0xf3},
    {0xcc, 0xfc},
    {0xcc, 0xff},
    {0xcf, 0x00},
    {0xcf, 0x03},
    {0xcf, 0x0c},
    {0xcf, 0x0f},
    {0xcf, 0x30},
    {0xcf, 0x33},
    {0xcf, 0x3c},
    {0xcf, 0x3f},
    {0xcf, 0xc0},
    {0xcf, 0xc3},
    {0xcf, 0xcc},
    {0xcf, 0xcf},
    {0xcf, 0xf0},
    {0xcf, 0xf3},
    {0xcf, 0xfc},
    {0xcf, 0xff},
    {0xf0, 0x00},
    {0xf0, 0x03},
    {0xf0, 0x0c},
    {0xf0, 0x0f},
    {0xf0, 0x30},
    {0xf0, 0x33},
    {0xf0, 0x3c},
    {0xf0, 0x3f},
    {0xf0, 0xc0},
    {0xf0, 0xc3},
    {0xf0, 0xcc},
    {0xf0, 0xcf},
    {0xf0, 0xf0},
    {0xf0, 0xf3},
    {0xf0, 0xfc},
    {0xf0, 0xff},
    {0xf3, 0x00},
    {0xf3, 0x03},
    {0xf3, 0x0c},
    {0xf3, 0x0f},
    {0xf3, 0x30},
    {0xf3, 0x33},
    {0xf3, 0x3c},
    {0xf3, 0x3f},
    {0xf3, 0xc0},
    {0xf3, 0xc3},
    {0xf3, 0xcc},
    {0xf3, 0xcf},
    {0xf3, 0xf0},
    {0xf3, 0xf3},
    {0xf3, 0xfc},
    {0xf3, 0xff},
    {0xfc, 0x00},
    {0xfc, 0x03},
    {0xfc, 0x0c},
    {0xfc, 0x0f},
    {0xfc, 0x30},
    {0xfc, 0x33},
    {0xfc, 0x3c},
    {0xfc, 0x3f},
    {0xfc, 0xc0},
    {0xfc, 0xc3},
    {0xfc, 0xcc},
    {0xfc, 0xcf},
    {0xfc, 0xf0},
    {0xfc, 0xf3},
    {0xfc, 0xfc},
    {0xfc, 0xff},
    {0xff, 0x00},
    {0xff, 0x03},
    {0xff, 0x0c},
    {0xff, 0x0f},
    {0xff, 0x30},
    {0xff, 0x33},
    {0xff, 0x3c},
    {0xff, 0x3f},
    {0xff, 0xc0},
    {0xff, 0xc3},
    {0xff, 0xcc},
    {0xff, 0xcf},
    {0xff, 0xf0},
    {0xff, 0xf3},
    {0xff, 0xfc},
    {0xff, 0xff}};    

  unsigned char *q = malloc(size*2*16);
  int i;

  assert(q != 0);
  for (i=0; i<size*8; i++) {
    memcpy(&q[4*i],conv[p[i]],2);
    memcpy(&q[4*i+2],conv[p[i]],2);
  }
  return q;
}

void d_write(char *p, int x, int y)
{
  gl_setfont(8,8,gl_font8x8);
  gl_write(PLAYGROUNDX+4*x, PLAYGROUNDY+4*y, p);
}

void d_demomessage()
{
  d_write(" JOYSTICK UP", 48, -2);
  d_write("   TO PLAY", 48, 0);
  d_write("JOYSTICK DOWN", 48, 4);
  d_write("   TO QUIT", 48, 6);
}

void d_democlear()
{
  gl_fillbox(PLAYGROUNDX+192, PLAYGROUNDY-8, 104, 40, 0);
}

void d_wrscore(int player, char *score)
{
  d_write(score, 0+36*player, -2);
}

void d_timedisplay(int player, char *time)
{
  d_write("TIME", 16, -2);
  d_write(time, 26, -2);
}

void d_wrtime(int player, char *time)
{
  d_write(time, 26, -2);
}

void d_timeclear()
{
  gl_fillbox(PLAYGROUNDX+64, PLAYGROUNDY-8, 64, 8, 0);
}
  
  
unsigned char *d_bigfont;

void d_wrbig(char *p, int x, int y)
{
  gl_setfont(16,16,d_bigfont);
  gl_write(PLAYGROUNDX+4*x, PLAYGROUNDY+4*y, p);
}

void d_init(void)
{
  static unsigned char palette[8][3] = {{0,0,0},
					{63,0,0},
					{0,63,0},
					{63,63,0},
					{0,0,63},
					{63,0,63},
					{0,63,63},
					{63,63,63} };
  /*  vga_safety_fork(0);*/
  vga_init();
  atexit(d_exit);
  vga_setmode(G320x200x256);
  gl_setpalettecolors(0,8,palette);
  gl_clearscreen(1);
  gl_setcontextvga(G320x200x256);
  gl_setpalettecolors(0,8,palette);
  gl_setfontcolors(0,7);
  gl_setwritemode(WRITEMODE_OVERWRITE|FONT_COMPRESSED);
  d_bigfont = d_makebigfont(gl_font8x8,256);
  if (keyboard_init() != 0) {
    fprintf(stderr,"Unable to access keyboard");
    exit(4);
  }
  atexit(keyboard_close);
  keyboard_translatekeys(TRANSLATE_CURSORKEYS);
}

unsigned d_joypause(int n, unsigned joymask)
{
  struct timeval to, *tp;
  fd_set junk;
  int i;
  unsigned js;

  FD_ZERO(&junk);
  if (n) {
    to.tv_usec = 0;
    to.tv_sec = n;
    tp = &to;
  } else
    tp = NULL;
  for (;;)
    {
      if ((js = d_joystc()) & joymask)
	return js;
      i = vga_waitevent(VGA_KEYEVENT, &junk, NULL, NULL, tp);
      if (tp == NULL)
	continue;
      if (i < 0 || (to.tv_usec == 0 && to.tv_sec == 0))
	return d_joystc();
    }
}
      
void d_sleep(struct timeval * tv)
{
  select(0,0,0,0,tv);
}

void d_askkey(void)
{
  keyboard_update();
}

int d_smarky(void)
{
  static int oldsmp = 0;
  int smp;

  d_askkey();
  smp = keyboard_keypressed(SCANCODE_SPACE);
  if (smp && !oldsmp) {
    oldsmp = 1;
    return 1;
  }
  oldsmp = smp;
  return 0;
}

int d_fireky(void)
{
  d_askkey();
  return keyboard_keypressed(SCANCODE_LEFTCONTROL);
}

unsigned d_joystc(void)
{
  char *p;
  int c;

  d_askkey();
  p = keyboard_getstate();
  c = p[SCANCODE_CURSORDOWN]<<3|p[SCANCODE_CURSORUP]<<2
    |p[SCANCODE_CURSORLEFT]<<1|p[SCANCODE_CURSORRIGHT];
  return c;
}
